@@ -290,6 +290,7 @@ enum ie_type {
 enum ie_vendor_wfa_oi_type {
 	IE_WFA_OI_HS20_INDICATION = 0x10,
 	IE_WFA_OI_OSEN = 0x12,
+	IE_WFA_OI_OWE_TRANSITION = 0x1c,
 	IE_WFA_OI_ROAMING_SELECTION = 0x1d,
 };
 
@@ -393,6 +394,7 @@ struct ie_rsn_info {
 	bool spp_a_msdu_required:1;
 	bool pbac:1;
 	bool extended_key_id:1;
+	bool ocvc:1;
 	uint8_t num_pmkids;
 	const uint8_t *pmkids;
 	enum ie_rsn_cipher_suite group_management_cipher;
@@ -417,6 +419,7 @@ enum ie_bss_capability {
 
 struct ie_ft_info {
 	uint8_t mic_element_count;
+	bool rsnxe_used : 1;
 	uint8_t mic[24];
 	uint8_t anonce[32];
 	uint8_t snonce[32];
@@ -432,6 +435,8 @@ struct ie_ft_info {
 	uint8_t igtk_ipn[6];
 	uint8_t igtk_len;
 	uint8_t igtk[24];
+	bool oci_present:1;
+	uint8_t oci[3];
 };
 
 /* See chapter 8.4.2.47 for radio measurement capability details */
@@ -459,6 +464,41 @@ struct ie_neighbor_report_info {
 	bool bss_transition_pref_present : 1;
 };
 
+struct ie_fils_ip_addr_request_info {
+	bool ipv4 : 1;
+	uint32_t ipv4_requested_addr;		/* Zero if none */
+	bool ipv6 : 1;
+	uint8_t ipv6_requested_addr[16];	/* Zero if none */
+	bool dns : 1;
+};
+
+struct ie_fils_ip_addr_response_info {
+	bool response_pending : 1;
+	uint8_t response_timeout;	/* Seconds */
+	uint32_t ipv4_addr;		/* Zero if not provided */
+	uint8_t ipv4_prefix_len;
+	uint32_t ipv4_gateway;		/* Zero if not provided */
+	uint8_t ipv4_gateway_mac[6];
+	uint32_t ipv4_dns;		/* Zero if not provided */
+	uint8_t ipv4_dns_mac[6];	/* Zero if not provided */
+	uint8_t ipv4_lifetime;		/* Zero if not provided */
+	uint8_t ipv6_addr[16];		/* Zero if not provided */
+	uint8_t ipv6_prefix_len;
+	uint8_t ipv6_gateway[16];	/* Zero if not provided */
+	uint8_t ipv6_gateway_mac[6];
+	uint8_t ipv6_dns[16];		/* Zero if not provided */
+	uint8_t ipv6_dns_mac[6];	/* Zero if not provided */
+	uint8_t ipv6_lifetime;		/* Zero if not provided */
+};
+
+struct ie_owe_transition_info {
+	uint8_t bssid[6];
+	uint8_t ssid[32];
+	size_t ssid_len;
+	uint8_t oper_class;
+	uint8_t channel;
+};
+
 extern const unsigned char ieee_oui[3];
 extern const unsigned char microsoft_oui[3];
 extern const unsigned char wifi_alliance_oui[3];
@@ -524,6 +564,8 @@ int ie_parse_wpa_from_data(const uint8_t *data, size_t len,
 						struct ie_rsn_info *info);
 bool is_ie_wfa_ie(const uint8_t *data, uint8_t len, uint8_t oi_type);
 bool is_ie_wpa_ie(const uint8_t *data, uint8_t len);
+bool is_ie_default_sae_group_oui(const uint8_t *data, uint16_t len);
+
 bool ie_build_wpa(const struct ie_rsn_info *info, uint8_t *to);
 
 int ie_parse_bss_load(struct ie_tlv_iter *iter, uint16_t *out_sta_count,
@@ -586,3 +628,22 @@ enum ie_rsnx_capability {
 };
 
 bool ie_rsnxe_capable(const uint8_t *rsnxe, unsigned int bit);
+
+int ie_parse_fils_ip_addr_request(struct ie_tlv_iter *iter,
+				struct ie_fils_ip_addr_request_info *out);
+void ie_build_fils_ip_addr_request(
+				const struct ie_fils_ip_addr_request_info *info,
+				uint8_t *to);
+int ie_parse_fils_ip_addr_response(struct ie_tlv_iter *iter,
+				struct ie_fils_ip_addr_response_info *out);
+void ie_build_fils_ip_addr_response(
+			const struct ie_fils_ip_addr_response_info *info,
+			uint8_t *to);
+
+int ie_parse_network_cost(const void *data, size_t len,
+				uint16_t *flags, uint16_t *level);
+
+int ie_parse_owe_transition(const void *data, size_t len,
+				struct ie_owe_transition_info *info);
+
+int ie_parse_oci(const void *data, size_t len, const uint8_t **oci);
