@@ -58,9 +58,17 @@ struct netconfig {
 	uint8_t rtm_protocol;
 	uint8_t rtm_v6_protocol;
 	struct l_rtnl_address *v4_address;
+	struct l_rtnl_address *v6_address;
 	char **dns4_overrides;
 	char **dns6_overrides;
+	char **dns4_list;
+	char **dns6_list;
+	char *mdns;
 	struct ie_fils_ip_addr_response_info *fils_override;
+	char *v4_gateway_str;
+	char *v6_gateway_str;
+	char *v4_domain;
+	char **v6_domains;
 
 	const struct l_settings *active_settings;
 
@@ -74,6 +82,7 @@ struct netconfig {
 	uint32_t addr4_add_cmd_id;
 	uint32_t addr6_add_cmd_id;
 	uint32_t route4_add_gateway_cmd_id;
+	uint32_t route6_add_cmd_id;
 };
 
 static struct l_netlink *rtnl;
@@ -125,6 +134,22 @@ static int sysfs_write_ipv6_setting(const char *ifname, const char *setting,
 	return r;
 }
 
+static void netconfig_free_settings(struct netconfig *netconfig)
+{
+	l_rtnl_address_free(netconfig->v4_address);
+	netconfig->v4_address = NULL;
+	l_rtnl_address_free(netconfig->v6_address);
+	netconfig->v6_address = NULL;
+
+	l_strfreev(netconfig->dns4_overrides);
+	netconfig->dns4_overrides = NULL;
+	l_strfreev(netconfig->dns6_overrides);
+	netconfig->dns6_overrides = NULL;
+
+	l_free(netconfig->mdns);
+	netconfig->mdns = NULL;
+}
+
 static void netconfig_free(void *data)
 {
 	struct netconfig *netconfig = data;
@@ -152,20 +177,6 @@ static struct netconfig *netconfig_find(uint32_t ifindex)
 	return NULL;
 }
 
-#define APPEND_STRDUPV(dest, index, src)			\
-	do {							\
-		char **p;					\
-		for (p = src; p && *p; p++)			\
-			dest[index++] = l_strdup(*p);		\
-	} while (0)						\
-
-#define APPENDV(dest, index, src)				\
-	do {							\
-		char **p;					\
-		for (p = src; p && *p; p++)			\
-			dest[index++] = *p;			\
-	} while (0)						\
-
 static inline char *netconfig_ipv4_to_string(uint32_t addr)
 {
 	struct in_addr in_addr = { .s_addr = addr };
@@ -185,7 +196,7 @@ static inline char *netconfig_ipv6_to_string(const uint8_t *addr)
 	struct in6_addr in6_addr;
 	char *addr_str = l_malloc(INET6_ADDRSTRLEN);
 
-	memcpy(in6_addr.__in6_u.__u6_addr8, addr, 16);
+	memcpy(in6_addr.s6_addr, addr, 16);
 
 	if (L_WARN_ON(unlikely(!inet_ntop(AF_INET6, &in6_addr, addr_str,
 						INET6_ADDRSTRLEN)))) {
@@ -196,6 +207,72 @@ static inline char *netconfig_ipv6_to_string(const uint8_t *addr)
 	return addr_str;
 }
 
+static char **netconfig_get_dns_list(struct netconfig *netconfig, int af,
+					const uint8_t **out_dns_mac)
+{
+	const struct ie_fils_ip_addr_response_info *fils =
+		netconfig->fils_override;
+
+	if (af == AF_INET) {
+		const struct l_dhcp_lease *lease;
+
+		if (netconfig->dns4_overrides)
+			return l_strv_copy(netconfig->dns4_overrides);
+
+		if (netconfig->rtm_protocol != RTPROT_DHCP)
+			return NULL;
+
+		if (fils && fils->ipv4_dns) {
+			char **dns_list = l_new(char *, 2);
+
+			if (!l_memeqzero(fils->ipv4_dns_mac, 6) &&
+					out_dns_mac &&
+					util_ip_subnet_match(
+							fils->ipv4_prefix_len,
+							&fils->ipv4_addr,
+							&fils->ipv4_dns))
+				*out_dns_mac = fils->ipv4_dns_mac;
+
+			dns_list[0] = netconfig_ipv4_to_string(fils->ipv4_dns);
+			return dns_list;
+		}
+
+		if (!(lease = l_dhcp_client_get_lease(netconfig->dhcp_client)))
+			return NULL;
+
+		return l_dhcp_lease_get_dns(lease);
+	} else {
+		const struct l_dhcp6_lease *lease;
+
+		if (netconfig->dns6_overrides)
+			return l_strv_copy(netconfig->dns6_overrides);
+
+		if (netconfig->rtm_v6_protocol != RTPROT_DHCP)
+			return NULL;
+
+		if (fils && !l_memeqzero(fils->ipv6_dns, 16)) {
+			char **dns_list = l_new(char *, 2);
+
+			if (!l_memeqzero(fils->ipv6_dns_mac, 6) &&
+					out_dns_mac &&
+					util_ip_subnet_match(
+							fils->ipv6_prefix_len,
+							fils->ipv6_addr,
+							fils->ipv6_dns))
+				*out_dns_mac = fils->ipv6_dns_mac;
+
+			dns_list[0] = netconfig_ipv6_to_string(fils->ipv6_dns);
+			return dns_list;
+		}
+
+		if (!(lease = l_dhcp6_client_get_lease(
+						netconfig->dhcp6_client)))
+			return NULL;
+
+		return l_dhcp6_lease_get_dns(lease);
+	}
+}
+
 static void netconfig_set_neighbor_entry_cb(int error,
 						uint16_t type, const void *data,
 						uint32_t len, void *user_data)
@@ -205,89 +282,61 @@ static void netconfig_set_neighbor_entry_cb(int error,
 			strerror(-error), error);
 }
 
-static int netconfig_set_dns(struct netconfig *netconfig)
+static void netconfig_set_dns(struct netconfig *netconfig)
 {
-	char **dns6_list = NULL;
-	char **dns4_list = NULL;
-	unsigned int n_entries = 0;
-	char **dns_list;
-	const uint8_t *fils_dns4_mac = NULL;
-	const uint8_t *fils_dns6_mac = NULL;
-	const struct ie_fils_ip_addr_response_info *fils =
-		netconfig->fils_override;
+	if (!netconfig->dns4_list && !netconfig->dns6_list)
+		return;
 
-	if (!netconfig->dns4_overrides &&
-			netconfig->rtm_protocol == RTPROT_DHCP) {
-		const struct l_dhcp_lease *lease;
+	if (netconfig->dns4_list && netconfig->dns6_list) {
+		unsigned int n_entries4 = l_strv_length(netconfig->dns4_list);
+		unsigned int n_entries6 = l_strv_length(netconfig->dns6_list);
+		char **dns_list = l_malloc(sizeof(char *) *
+					(n_entries4 + n_entries6 + 1));
 
-		if (fils && fils->ipv4_dns) {
-			dns4_list = l_new(char *, 2);
-			dns4_list[0] = netconfig_ipv4_to_string(fils->ipv4_dns);
-
-			if (!l_memeqzero(fils->ipv4_dns_mac, 6) &&
-					util_ip_subnet_match(
-							fils->ipv4_prefix_len,
-							&fils->ipv4_addr,
-							&fils->ipv4_dns))
-				fils_dns4_mac = fils->ipv4_dns_mac;
-		} else if ((lease = l_dhcp_client_get_lease(
-						netconfig->dhcp_client)))
-			dns4_list = l_dhcp_lease_get_dns(lease);
+		memcpy(dns_list, netconfig->dns4_list,
+			sizeof(char *) * n_entries4);
+		memcpy(dns_list + n_entries4, netconfig->dns6_list,
+			sizeof(char *) * (n_entries6 + 1));
+		resolve_set_dns(netconfig->resolve, dns_list);
+		l_free(dns_list);
+		return;
 	}
 
-	if (!netconfig->dns6_overrides &&
-			netconfig->rtm_v6_protocol == RTPROT_DHCP) {
-		const struct l_dhcp6_lease *lease;
+	resolve_set_dns(netconfig->resolve,
+			netconfig->dns4_list ?: netconfig->dns6_list);
+}
 
-		if (fils && !l_memeqzero(fils->ipv6_dns,
-						16)) {
-			dns6_list = l_new(char *, 2);
-			dns6_list[0] = netconfig_ipv6_to_string(fils->ipv6_dns);
+static bool netconfig_dns_list_update(struct netconfig *netconfig, uint8_t af)
+{
+	const uint8_t *fils_dns_mac = NULL;
+	char ***dns_list_ptr = af == AF_INET ?
+		&netconfig->dns4_list : &netconfig->dns6_list;
+	char **new_dns_list = netconfig_get_dns_list(netconfig, af,
+							&fils_dns_mac);
 
-			if (!l_memeqzero(fils->ipv6_dns_mac, 6) &&
-					util_ip_subnet_match(
-							fils->ipv6_prefix_len,
-							fils->ipv6_addr,
-							fils->ipv6_dns))
-				fils_dns6_mac = fils->ipv6_dns_mac;
-		} else if ((lease = l_dhcp6_client_get_lease(
-						netconfig->dhcp6_client)))
-			dns6_list = l_dhcp6_lease_get_dns(lease);
+	if (l_strv_eq(*dns_list_ptr, new_dns_list)) {
+		l_strv_free(new_dns_list);
+		return false;
 	}
 
-	n_entries += l_strv_length(netconfig->dns4_overrides);
-	n_entries += l_strv_length(netconfig->dns6_overrides);
-	n_entries += l_strv_length(dns4_list);
-	n_entries += l_strv_length(dns6_list);
+	l_strv_free(*dns_list_ptr);
+	*dns_list_ptr = new_dns_list;
 
-	dns_list = l_new(char *, n_entries + 1);
-	n_entries = 0;
+	if (fils_dns_mac) {
+		const struct ie_fils_ip_addr_response_info *fils =
+			netconfig->fils_override;
+		const void *dns_ip = af == AF_INET ?
+			(const void *) &fils->ipv4_dns :
+			(const void *) &fils->ipv6_dns;
 
-	APPEND_STRDUPV(dns_list, n_entries, netconfig->dns4_overrides);
-	APPENDV(dns_list, n_entries, dns4_list);
-	APPEND_STRDUPV(dns_list, n_entries, netconfig->dns6_overrides);
-	APPENDV(dns_list, n_entries, dns6_list);
-	resolve_set_dns(netconfig->resolve, dns_list);
+		if (!l_rtnl_neighbor_set_hwaddr(rtnl, netconfig->ifindex, af,
+						dns_ip, fils_dns_mac, 6,
+						netconfig_set_neighbor_entry_cb,
+						NULL, NULL))
+			l_debug("l_rtnl_neighbor_set_hwaddr failed");
+	}
 
-	if (fils_dns4_mac && !l_rtnl_neighbor_set_hwaddr(rtnl,
-					netconfig->ifindex, AF_INET,
-					&fils->ipv4_dns, fils_dns4_mac, 6,
-					netconfig_set_neighbor_entry_cb, NULL,
-					NULL))
-		l_debug("l_rtnl_neighbor_set_hwaddr failed");
-
-	if (fils_dns6_mac && !l_rtnl_neighbor_set_hwaddr(rtnl,
-					netconfig->ifindex, AF_INET6,
-					fils->ipv6_dns, fils_dns6_mac, 6,
-					netconfig_set_neighbor_entry_cb, NULL,
-					NULL))
-		l_debug("l_rtnl_neighbor_set_hwaddr failed");
-
-	l_strv_free(dns_list);
-	/* Contents belonged to dns_list, so not l_strfreev */
-	l_free(dns4_list);
-	l_free(dns6_list);
-	return 0;
+	return true;
 }
 
 static void append_domain(char **domains, unsigned int *n_domains,
@@ -306,47 +355,71 @@ static void append_domain(char **domains, unsigned int *n_domains,
 	*n_domains += 1;
 }
 
-static int netconfig_set_domains(struct netconfig *netconfig)
+static void netconfig_set_domains(struct netconfig *netconfig)
 {
 	char *domains[31];
 	unsigned int n_domains = 0;
-	char *v4_domain = NULL;
-	char **v6_domains = NULL;
 	char **p;
 
 	memset(domains, 0, sizeof(domains));
 
-	/* Allow to override the DHCP domain name with setting entry. */
-	v4_domain = l_settings_get_string(netconfig->active_settings,
-							"IPv4", "DomainName");
-	if (!v4_domain && netconfig->rtm_protocol == RTPROT_DHCP) {
-		const struct l_dhcp_lease *lease =
-			l_dhcp_client_get_lease(netconfig->dhcp_client);
-
-		if (lease)
-			v4_domain = l_dhcp_lease_get_domain_name(lease);
-	}
-
 	append_domain(domains, &n_domains,
-					L_ARRAY_SIZE(domains) - 1, v4_domain);
+			L_ARRAY_SIZE(domains) - 1, netconfig->v4_domain);
 
-	if (netconfig->rtm_v6_protocol == RTPROT_DHCP) {
-		const struct l_dhcp6_lease *lease =
-			l_dhcp6_client_get_lease(netconfig->dhcp6_client);
-
-		if (lease)
-			v6_domains = l_dhcp6_lease_get_domains(lease);
-	}
-
-	for (p = v6_domains; p && *p; p++)
+	for (p = netconfig->v6_domains; p && *p; p++)
 		append_domain(domains, &n_domains,
-					L_ARRAY_SIZE(domains) - 1, *p);
+				L_ARRAY_SIZE(domains) - 1, *p);
 
 	resolve_set_domains(netconfig->resolve, domains);
-	l_strv_free(v6_domains);
-	l_free(v4_domain);
+}
 
-	return 0;
+static bool netconfig_domains_update(struct netconfig *netconfig, uint8_t af)
+{
+	bool changed = false;
+
+	if (af == AF_INET) {
+		/* Allow to override the DHCP domain name with setting entry. */
+		char *v4_domain = l_settings_get_string(
+						netconfig->active_settings,
+						"IPv4", "DomainName");
+
+		if (!v4_domain && netconfig->rtm_protocol == RTPROT_DHCP) {
+			const struct l_dhcp_lease *lease =
+				l_dhcp_client_get_lease(netconfig->dhcp_client);
+
+			if (lease)
+				v4_domain = l_dhcp_lease_get_domain_name(lease);
+		}
+
+		if (l_streq0(v4_domain, netconfig->v4_domain))
+			l_free(v4_domain);
+		else {
+			l_free(netconfig->v4_domain);
+			netconfig->v4_domain = v4_domain;
+			changed = true;
+		}
+	} else {
+		char **v6_domains = NULL;
+
+		if (netconfig->rtm_v6_protocol == RTPROT_DHCP) {
+			const struct l_dhcp6_lease *lease =
+				l_dhcp6_client_get_lease(
+						netconfig->dhcp6_client);
+
+			if (lease)
+				v6_domains = l_dhcp6_lease_get_domains(lease);
+		}
+
+		if (l_strv_eq(netconfig->v6_domains, v6_domains))
+			l_strv_free(v6_domains);
+		else {
+			l_strv_free(netconfig->v6_domains);
+			netconfig->v6_domains = v6_domains;
+			changed = true;
+		}
+	}
+
+	return changed;
 }
 
 static struct l_rtnl_address *netconfig_get_static4_address(
@@ -402,6 +475,8 @@ static char *netconfig_ipv4_get_gateway(struct netconfig *netconfig,
 {
 	const struct l_dhcp_lease *lease;
 	char *gateway;
+	const struct ie_fils_ip_addr_response_info *fils =
+		netconfig->fils_override;
 
 	switch (netconfig->rtm_protocol) {
 	case RTPROT_STATIC:
@@ -415,15 +490,12 @@ static char *netconfig_ipv4_get_gateway(struct netconfig *netconfig,
 		return gateway;
 
 	case RTPROT_DHCP:
-		if (netconfig->fils_override &&
-				netconfig->fils_override->ipv4_gateway) {
-			gateway = netconfig_ipv4_to_string(
-					netconfig->fils_override->ipv4_gateway);
+		if (fils && fils->ipv4_gateway) {
+			gateway = netconfig_ipv4_to_string(fils->ipv4_gateway);
 
-			if (gateway && !l_memeqzero(netconfig->fils_override->
-							ipv4_gateway_mac, 6))
-				*out_mac = netconfig->fils_override->
-					ipv4_gateway_mac;
+			if (gateway && out_mac &&
+					!l_memeqzero(fils->ipv4_gateway_mac, 6))
+				*out_mac = fils->ipv4_gateway_mac;
 
 			return gateway;
 		}
@@ -461,7 +533,7 @@ static struct l_rtnl_address *netconfig_get_static6_address(
 	prefix_len = strtoul(p, NULL, 10);
 	if (!unlikely(errno == EINVAL || errno == ERANGE ||
 			!prefix_len || prefix_len > 128)) {
-		l_error("netconfig: Invalid prefix '%s'  provided in network"
+		l_error("netconfig: Invalid prefix '%s' provided in network"
 				" configuration file", p);
 		return NULL;
 	}
@@ -477,10 +549,12 @@ no_prefix_len:
 
 static struct l_rtnl_route *netconfig_get_static6_gateway(
 						struct netconfig *netconfig,
+						char **out_str,
 						const uint8_t **out_mac)
 {
 	L_AUTO_FREE_VAR(char *, gateway);
 	struct l_rtnl_route *ret;
+	const uint8_t *mac = NULL;
 
 	gateway = l_settings_get_string(netconfig->active_settings,
 						"IPv6", "Gateway");
@@ -492,7 +566,7 @@ static struct l_rtnl_route *netconfig_get_static6_gateway(
 					netconfig->fils_override->ipv6_gateway);
 
 		if (!l_memeqzero(netconfig->fils_override->ipv6_gateway_mac, 6))
-			*out_mac = netconfig->fils_override->ipv6_gateway_mac;
+			mac = netconfig->fils_override->ipv6_gateway_mac;
 	} else if (!gateway)
 		return NULL;
 
@@ -506,6 +580,8 @@ static struct l_rtnl_route *netconfig_get_static6_gateway(
 
 	l_rtnl_route_set_priority(ret, ROUTE_PRIORITY_OFFSET);
 	l_rtnl_route_set_protocol(ret, RTPROT_STATIC);
+	*out_str = l_steal_ptr(gateway);
+	*out_mac = mac;
 
 	return ret;
 }
@@ -516,9 +592,7 @@ static struct l_rtnl_address *netconfig_get_dhcp4_address(
 	const struct l_dhcp_lease *lease =
 			l_dhcp_client_get_lease(netconfig->dhcp_client);
 	L_AUTO_FREE_VAR(char *, ip) = NULL;
-	L_AUTO_FREE_VAR(char *, netmask) = NULL;
 	L_AUTO_FREE_VAR(char *, broadcast) = NULL;
-	struct in_addr in_addr;
 	uint32_t prefix_len;
 	struct l_rtnl_address *ret;
 
@@ -526,12 +600,10 @@ static struct l_rtnl_address *netconfig_get_dhcp4_address(
 		return NULL;
 
 	ip = l_dhcp_lease_get_address(lease);
-	netmask = l_dhcp_lease_get_netmask(lease);
 	broadcast = l_dhcp_lease_get_broadcast(lease);
 
-	if (netmask && inet_pton(AF_INET, netmask, &in_addr) > 0)
-		prefix_len = __builtin_popcountl(in_addr.s_addr);
-	else
+	prefix_len = l_dhcp_lease_get_prefix_length(lease);
+	if (!prefix_len)
 		prefix_len = 24;
 
 	ret = l_rtnl_address_new(ip, prefix_len);
@@ -545,6 +617,39 @@ static struct l_rtnl_address *netconfig_get_dhcp4_address(
 	return ret;
 }
 
+static void netconfig_gateway_to_arp(struct netconfig *netconfig)
+{
+	const struct l_dhcp_lease *lease;
+	_auto_(l_free) char *server_id = NULL;
+	_auto_(l_free) char *gw = NULL;
+	const uint8_t *server_mac;
+	struct in_addr in_gw;
+
+	/* Can only do this for DHCP in certain network setups */
+	if (netconfig->rtm_protocol != RTPROT_DHCP)
+		return;
+
+	lease = l_dhcp_client_get_lease(netconfig->dhcp_client);
+	if (!lease)
+		return;
+
+	server_id = l_dhcp_lease_get_server_id(lease);
+	gw = l_dhcp_lease_get_gateway(lease);
+	server_mac = l_dhcp_lease_get_server_mac(lease);
+
+	if (!gw || strcmp(server_id, gw) || !server_mac)
+		return;
+
+	l_debug("Gateway MAC is known, setting into ARP cache");
+	in_gw.s_addr = l_dhcp_lease_get_gateway_u32(lease);
+
+	if (!l_rtnl_neighbor_set_hwaddr(rtnl, netconfig->ifindex, AF_INET,
+					&in_gw, server_mac, ETH_ALEN,
+					netconfig_set_neighbor_entry_cb, NULL,
+					NULL))
+		l_debug("l_rtnl_neighbor_set_hwaddr failed");
+}
+
 static void netconfig_ifaddr_added(struct netconfig *netconfig,
 					const struct ifaddrmsg *ifa,
 					uint32_t len)
@@ -724,6 +829,21 @@ static void netconfig_route_add_cmd_cb(int error, uint16_t type,
 	netconfig->notify = NULL;
 }
 
+static void netconfig_route6_add_cb(int error, uint16_t type,
+					const void *data, uint32_t len,
+					void *user_data)
+{
+	struct netconfig *netconfig = user_data;
+
+	netconfig->route6_add_cmd_id = 0;
+
+	if (error) {
+		l_error("netconfig: Failed to add route. Error %d: %s",
+						error, strerror(-error));
+		return;
+	}
+}
+
 static bool netconfig_ipv4_routes_install(struct netconfig *netconfig)
 {
 	L_AUTO_FREE_VAR(char *, gateway) = NULL;
@@ -817,6 +937,8 @@ static void netconfig_ipv4_ifaddr_add_cmd_cb(int error, uint16_t type,
 		return;
 	}
 
+	netconfig_gateway_to_arp(netconfig);
+
 	if (!netconfig_ipv4_routes_install(netconfig)) {
 		l_error("netconfig: Failed to install IPv4 routes.");
 		return;
@@ -832,7 +954,7 @@ static void netconfig_ipv6_ifaddr_add_cmd_cb(int error, uint16_t type,
 {
 	struct netconfig *netconfig = user_data;
 	struct l_rtnl_route *gateway;
-	const uint8_t *gateway_mac = NULL;
+	const uint8_t *gateway_mac;
 
 	netconfig->addr6_add_cmd_id = 0;
 
@@ -842,12 +964,16 @@ static void netconfig_ipv6_ifaddr_add_cmd_cb(int error, uint16_t type,
 		return;
 	}
 
-	gateway = netconfig_get_static6_gateway(netconfig, &gateway_mac);
+	gateway = netconfig_get_static6_gateway(netconfig,
+						&netconfig->v6_gateway_str,
+						&gateway_mac);
 	if (gateway) {
-		L_WARN_ON(!l_rtnl_route_add(rtnl, netconfig->ifindex,
-						gateway,
-						netconfig_route_generic_cb,
-						netconfig, NULL));
+		netconfig->route6_add_cmd_id = l_rtnl_route_add(rtnl,
+							netconfig->ifindex,
+							gateway,
+							netconfig_route6_add_cb,
+							netconfig, NULL);
+		L_WARN_ON(unlikely(!netconfig->route6_add_cmd_id));
 		l_rtnl_route_free(gateway);
 
 		if (gateway_mac && !l_rtnl_neighbor_set_hwaddr(rtnl,
@@ -897,22 +1023,40 @@ static void netconfig_ipv4_dhcp_event_handler(struct l_dhcp_client *client,
 					netconfig->v4_address,
 					netconfig_ifaddr_del_cmd_cb,
 					netconfig, NULL));
-		l_rtnl_address_free(netconfig->v4_address);
 		/* Fall through. */
 	case L_DHCP_CLIENT_EVENT_LEASE_OBTAINED:
-		netconfig->v4_address = netconfig_get_dhcp4_address(netconfig);
+	{
+		char *gateway_str;
+		struct l_rtnl_address *address;
+
+		gateway_str = netconfig_ipv4_get_gateway(netconfig, NULL);
+		if (l_streq0(netconfig->v4_gateway_str, gateway_str))
+			l_free(gateway_str);
+		else {
+			l_free(netconfig->v4_gateway_str);
+			netconfig->v4_gateway_str = gateway_str;
+		}
+
+		address = netconfig_get_dhcp4_address(netconfig);
+		l_rtnl_address_free(netconfig->v4_address);
+		netconfig->v4_address = address;
+
 		if (!netconfig->v4_address) {
 			l_error("netconfig: Failed to obtain IP addresses from "
 							"DHCPv4 lease.");
 			return;
 		}
 
+		netconfig_dns_list_update(netconfig, AF_INET);
+		netconfig_domains_update(netconfig, AF_INET);
+
 		L_WARN_ON(!(netconfig->addr4_add_cmd_id =
 				l_rtnl_ifaddr_add(rtnl, netconfig->ifindex,
 					netconfig->v4_address,
 					netconfig_ipv4_ifaddr_add_cmd_cb,
 					netconfig, NULL)));
 		break;
+	}
 	case L_DHCP_CLIENT_EVENT_LEASE_RENEWED:
 		break;
 	case L_DHCP_CLIENT_EVENT_LEASE_EXPIRED:
@@ -922,6 +1066,7 @@ static void netconfig_ipv4_dhcp_event_handler(struct l_dhcp_client *client,
 					netconfig, NULL));
 		l_rtnl_address_free(netconfig->v4_address);
 		netconfig->v4_address = NULL;
+		l_free(l_steal_ptr(netconfig->v4_gateway_str));
 
 		/* Fall through. */
 	case L_DHCP_CLIENT_EVENT_NO_LEASE:
@@ -950,13 +1095,45 @@ static void netconfig_dhcp6_event_handler(struct l_dhcp6_client *client,
 	case L_DHCP6_CLIENT_EVENT_IP_CHANGED:
 	case L_DHCP6_CLIENT_EVENT_LEASE_OBTAINED:
 	case L_DHCP6_CLIENT_EVENT_LEASE_RENEWED:
+	{
+		const struct l_dhcp6_lease *lease =
+			l_dhcp6_client_get_lease(netconfig->dhcp6_client);
+		_auto_(l_free) char *addr_str =
+			l_dhcp6_lease_get_address(lease);
+		struct l_rtnl_address *address;
+		struct l_icmp6_client *icmp6 =
+			l_dhcp6_client_get_icmp6(netconfig->dhcp6_client);
+		const struct l_icmp6_router *router =
+			l_icmp6_client_get_router(icmp6);
+		char *gateway_str = l_icmp6_router_get_address(router);
+
+		if (l_streq0(netconfig->v6_gateway_str, gateway_str))
+			l_free(gateway_str);
+		else {
+			l_free(netconfig->v6_gateway_str);
+			netconfig->v6_gateway_str = gateway_str;
+		}
+
+		address = l_rtnl_address_new(addr_str,
+					l_dhcp6_lease_get_prefix_length(lease));
+		l_rtnl_address_free(netconfig->v6_address);
+		netconfig->v6_address = address;
+
+		netconfig_dns_list_update(netconfig, AF_INET6);
+		netconfig_domains_update(netconfig, AF_INET6);
 		netconfig_set_dns(netconfig);
 		netconfig_set_domains(netconfig);
 		break;
+	}
 	case L_DHCP6_CLIENT_EVENT_LEASE_EXPIRED:
 		l_debug("Lease for interface %u expired", netconfig->ifindex);
+		netconfig_dns_list_update(netconfig, AF_INET6);
+		netconfig_domains_update(netconfig, AF_INET6);
 		netconfig_set_dns(netconfig);
 		netconfig_set_domains(netconfig);
+		l_rtnl_address_free(netconfig->v6_address);
+		netconfig->v6_address = NULL;
+		l_free(l_steal_ptr(netconfig->v6_gateway_str));
 
 		/* Fall through */
 	case L_DHCP6_CLIENT_EVENT_NO_LEASE:
@@ -985,14 +1162,18 @@ static void netconfig_reset_v4(struct netconfig *netconfig)
 	if (netconfig->rtm_protocol) {
 		netconfig_remove_v4_address(netconfig);
 
-		l_strfreev(netconfig->dns4_overrides);
-		netconfig->dns4_overrides = NULL;
+		l_strv_free(l_steal_ptr(netconfig->dns4_overrides));
+		l_strv_free(l_steal_ptr(netconfig->dns4_list));
 
 		l_dhcp_client_stop(netconfig->dhcp_client);
 		netconfig->rtm_protocol = 0;
 
 		l_acd_destroy(netconfig->acd);
 		netconfig->acd = NULL;
+
+		l_free(l_steal_ptr(netconfig->v4_gateway_str));
+
+		l_free(l_steal_ptr(netconfig->v4_domain));
 	}
 }
 
@@ -1031,6 +1212,7 @@ static void netconfig_ipv4_acd_event(enum l_acd_event event, void *user_data)
 
 static bool netconfig_ipv4_select_and_install(struct netconfig *netconfig)
 {
+	struct netdev *netdev = netdev_find(netconfig->ifindex);
 	bool set_address = (netconfig->rtm_protocol == RTPROT_STATIC);
 
 	if (netconfig->rtm_protocol == RTPROT_DHCP &&
@@ -1068,6 +1250,9 @@ static bool netconfig_ipv4_select_and_install(struct netconfig *netconfig)
 							ip)))
 			return false;
 
+		netconfig_dns_list_update(netconfig, AF_INET);
+		netconfig_domains_update(netconfig, AF_INET);
+
 		netconfig->acd = l_acd_new(netconfig->ifindex);
 		l_acd_set_event_handler(netconfig->acd,
 					netconfig_ipv4_acd_event, netconfig,
@@ -1091,6 +1276,9 @@ static bool netconfig_ipv4_select_and_install(struct netconfig *netconfig)
 		return true;
 	}
 
+	l_dhcp_client_set_address(netconfig->dhcp_client, ARPHRD_ETHER,
+					netdev_get_address(netdev), ETH_ALEN);
+
 	if (l_dhcp_client_start(netconfig->dhcp_client))
 		return true;
 
@@ -1102,7 +1290,6 @@ static bool netconfig_ipv4_select_and_install(struct netconfig *netconfig)
 static bool netconfig_ipv6_select_and_install(struct netconfig *netconfig)
 {
 	struct netdev *netdev = netdev_find(netconfig->ifindex);
-	struct l_rtnl_address *address = NULL;
 
 	if (netconfig->rtm_v6_protocol == RTPROT_UNSPEC) {
 		l_debug("IPV6 configuration disabled");
@@ -1111,10 +1298,7 @@ static bool netconfig_ipv6_select_and_install(struct netconfig *netconfig)
 
 	sysfs_write_ipv6_setting(netdev_get_name(netdev), "disable_ipv6", "0");
 
-	if (netconfig->rtm_v6_protocol == RTPROT_STATIC)
-		address = netconfig_get_static6_address(
-						netconfig->active_settings);
-	else if (netconfig->rtm_v6_protocol == RTPROT_DHCP &&
+	if (netconfig->rtm_v6_protocol == RTPROT_DHCP &&
 			netconfig->fils_override &&
 			!l_memeqzero(netconfig->fils_override->ipv6_addr, 16)) {
 		uint8_t prefix_len = netconfig->fils_override->ipv6_prefix_len;
@@ -1124,11 +1308,12 @@ static bool netconfig_ipv6_select_and_install(struct netconfig *netconfig)
 		if (unlikely(!addr_str))
 			return false;
 
-		if (L_WARN_ON(unlikely(!(address = l_rtnl_address_new(addr_str,
-								prefix_len)))))
+		netconfig->v6_address = l_rtnl_address_new(addr_str,
+								prefix_len);
+		if (L_WARN_ON(unlikely(!netconfig->v6_address)))
 			return false;
 
-		l_rtnl_address_set_noprefixroute(address, true);
+		l_rtnl_address_set_noprefixroute(netconfig->v6_address, true);
 
 		/*
 		 * TODO: If netconfig->fils_override->ipv6_lifetime is set,
@@ -1138,16 +1323,21 @@ static bool netconfig_ipv6_select_and_install(struct netconfig *netconfig)
 		 */
 	}
 
-	if (address) {
+	if (netconfig->v6_address) {
+		netconfig_dns_list_update(netconfig, AF_INET6);
+
 		L_WARN_ON(!(netconfig->addr6_add_cmd_id =
-			l_rtnl_ifaddr_add(rtnl, netconfig->ifindex, address,
+			l_rtnl_ifaddr_add(rtnl, netconfig->ifindex,
+					netconfig->v6_address,
 					netconfig_ipv6_ifaddr_add_cmd_cb,
 					netconfig, NULL)));
-		l_rtnl_address_free(address);
 		return true;
 	}
 
-	/* DHCP */
+	/* DHCPv6 or RA, update MAC */
+	l_dhcp6_client_set_address(netconfig->dhcp6_client, ARPHRD_ETHER,
+					netdev_get_address(netdev), ETH_ALEN);
+
 	return true;
 }
 
@@ -1181,17 +1371,16 @@ static int validate_dns_list(int family, char **dns_list)
 }
 
 bool netconfig_load_settings(struct netconfig *netconfig,
-				const struct l_settings *active_settings,
-				const uint8_t *mac_address)
+				const struct l_settings *active_settings)
 {
-	char *mdns;
+	_auto_(l_free) char *mdns = NULL;
 	bool send_hostname;
 	bool v6_enabled;
 	char hostname[HOST_NAME_MAX + 1];
-	char **dns4_overrides = NULL;
-	char **dns6_overrides = NULL;
-	struct l_rtnl_address *v4_address = NULL;
-	struct l_rtnl_address *v6_address = NULL;
+	_auto_(l_strv_free) char **dns4_overrides = NULL;
+	_auto_(l_strv_free) char **dns6_overrides = NULL;
+	_auto_(l_rtnl_address_free) struct l_rtnl_address *v4_address = NULL;
+	_auto_(l_rtnl_address_free) struct l_rtnl_address *v6_address = NULL;
 
 	dns4_overrides = l_settings_get_string_list(active_settings,
 							"IPv4", "DNS", ' ');
@@ -1203,7 +1392,7 @@ bool netconfig_load_settings(struct netconfig *netconfig,
 			dns4_overrides = NULL;
 
 			if (r < 0)
-				goto err_dns4;
+				return false;
 		}
 
 		if (r == 0)
@@ -1221,7 +1410,7 @@ bool netconfig_load_settings(struct netconfig *netconfig,
 			dns6_overrides = NULL;
 
 			if (r < 0)
-				goto err_dns6;
+				return false;
 		}
 
 		if (r == 0)
@@ -1248,7 +1437,7 @@ bool netconfig_load_settings(struct netconfig *netconfig,
 
 		if (unlikely(!v4_address)) {
 			l_error("netconfig: Can't parse IPv4 address");
-			goto err_v4_addr;
+			return false;
 		}
 	}
 
@@ -1258,21 +1447,15 @@ bool netconfig_load_settings(struct netconfig *netconfig,
 
 	if (l_settings_has_key(active_settings, "IPv6", "Address")) {
 		v6_address = netconfig_get_static6_address(active_settings);
-		l_rtnl_address_free(v6_address);
 
 		if (unlikely(!v6_address)) {
 			l_error("netconfig: Can't parse IPv6 address");
-			goto err_v6_addr;
+			return false;
 		}
 	}
 
 	/* No more validation steps for now, commit new values */
-
-	if (v4_address) {
-		netconfig->v4_address = v4_address;
-		netconfig->rtm_protocol = RTPROT_STATIC;
-	} else
-		netconfig->rtm_protocol = RTPROT_DHCP;
+	netconfig->rtm_protocol = v4_address ? RTPROT_STATIC : RTPROT_DHCP;
 
 	if (!v6_enabled)
 		netconfig->rtm_v6_protocol = RTPROT_UNSPEC;
@@ -1284,28 +1467,19 @@ bool netconfig_load_settings(struct netconfig *netconfig,
 	if (send_hostname)
 		l_dhcp_client_set_hostname(netconfig->dhcp_client, hostname);
 
-	resolve_set_mdns(netconfig->resolve, mdns);
-	l_free(mdns);
+	netconfig_free_settings(netconfig);
 
-	l_dhcp_client_set_address(netconfig->dhcp_client, ARPHRD_ETHER,
-							mac_address, ETH_ALEN);
-	l_dhcp6_client_set_address(netconfig->dhcp6_client, ARPHRD_ETHER,
-							mac_address, ETH_ALEN);
+	if (netconfig->rtm_protocol == RTPROT_STATIC)
+		netconfig->v4_address = l_steal_ptr(v4_address);
+
+	if (netconfig->rtm_v6_protocol == RTPROT_STATIC)
+		netconfig->v6_address = l_steal_ptr(v6_address);
 
 	netconfig->active_settings = active_settings;
-	netconfig->dns4_overrides = dns4_overrides;
-	netconfig->dns6_overrides = dns6_overrides;
+	netconfig->dns4_overrides = l_steal_ptr(dns4_overrides);
+	netconfig->dns6_overrides = l_steal_ptr(dns6_overrides);
+	netconfig->mdns = l_steal_ptr(mdns);
 	return true;
-
-err_v6_addr:
-	l_rtnl_address_free(v4_address);
-err_v4_addr:
-	l_free(mdns);
-	l_strfreev(dns6_overrides);
-err_dns6:
-	l_strfreev(dns4_overrides);
-err_dns4:
-	return false;
 }
 
 bool netconfig_configure(struct netconfig *netconfig,
@@ -1320,11 +1494,22 @@ bool netconfig_configure(struct netconfig *netconfig,
 	if (unlikely(!netconfig_ipv6_select_and_install(netconfig)))
 		return false;
 
+	resolve_set_mdns(netconfig->resolve, netconfig->mdns);
+
 	return true;
 }
 
 bool netconfig_reconfigure(struct netconfig *netconfig)
 {
+	/*
+	 * Starting with kernel 4.20, ARP cache is flushed when the netdev
+	 * detects NO CARRIER.  This can result in unnecessarily long delays
+	 * (about 1 second on some networks) due to ARP query response being
+	 * lost or delayed.  Try to force the gateway into the ARP cache
+	 * to alleviate this
+	 */
+	netconfig_gateway_to_arp(netconfig);
+
 	if (netconfig->rtm_protocol == RTPROT_DHCP) {
 		/* TODO l_dhcp_client sending a DHCP inform request */
 	}
@@ -1345,6 +1530,11 @@ bool netconfig_reset(struct netconfig *netconfig)
 		netconfig->route4_add_gateway_cmd_id = 0;
 	}
 
+	if (netconfig->route6_add_cmd_id) {
+		l_netlink_cancel(rtnl, netconfig->route6_add_cmd_id);
+		netconfig->route6_add_cmd_id = 0;
+	}
+
 	if (netconfig->addr4_add_cmd_id) {
 		l_netlink_cancel(rtnl, netconfig->addr4_add_cmd_id);
 		netconfig->addr4_add_cmd_id = 0;
@@ -1361,14 +1551,21 @@ bool netconfig_reset(struct netconfig *netconfig)
 	netconfig_reset_v4(netconfig);
 
 	if (netconfig->rtm_v6_protocol) {
-		l_strfreev(netconfig->dns6_overrides);
-		netconfig->dns6_overrides = NULL;
+		l_rtnl_address_free(netconfig->v6_address);
+		netconfig->v6_address = NULL;
+
+		l_strv_free(l_steal_ptr(netconfig->dns6_overrides));
+		l_strv_free(l_steal_ptr(netconfig->dns6_list));
 
 		l_dhcp6_client_stop(netconfig->dhcp6_client);
 		netconfig->rtm_v6_protocol = 0;
 
 		sysfs_write_ipv6_setting(netdev_get_name(netdev),
 						"disable_ipv6", "1");
+
+		l_free(l_steal_ptr(netconfig->v6_gateway_str));
+
+		l_strv_free(l_steal_ptr(netconfig->v6_domains));
 	}
 
 	l_free(l_steal_ptr(netconfig->fils_override));
@@ -1485,6 +1682,15 @@ void netconfig_destroy(struct netconfig *netconfig)
 	netconfig_free(netconfig);
 }
 
+bool netconfig_enabled(void)
+{
+	bool enabled;
+
+	return l_settings_get_bool(iwd_get_config(), "General",
+					"EnableNetworkConfiguration",
+					&enabled) && enabled;
+}
+
 static int netconfig_init(void)
 {
 	uint32_t r;
